import shutil
import unittest
from io import StringIO
from unittest.mock import patch

from pocketbook.disclaimer import display_disclaimer, DISCLAIMER
from .utils import TemporaryPocketBookRoot


class DisclaimerTests(unittest.TestCase):

    def test_display_of_disclaimer(self):
        with TemporaryPocketBookRoot() as ctx:
            shutil.rmtree(ctx.root)  # remove the root - emulate initial startup

            with patch('builtins.input', return_value='\n'):
                with patch('sys.stdout', new_callable=StringIO) as mocked_print:
                    display_disclaimer(ctx.root)
                    self.assertIn(DISCLAIMER, mocked_print.getvalue())

    def test_no_display_of_disclaimer(self):
        with TemporaryPocketBookRoot() as ctx:
            with patch('builtins.input', return_value='\n'):
                with patch('sys.stdout', new_callable=StringIO) as mocked_print:
                    display_disclaimer(ctx.root)
                    self.assertEqual('', mocked_print.getvalue())



import unittest
from unittest.mock import patch, Mock

from pocketbook.address_book import AddressBook
from pocketbook.key_store import KeyStore


class RenameCommandTests(unittest.TestCase):

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_normal_key_rename(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['old-name']

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 0)

        key_store.rename_key.assert_called_once_with('old-name', 'new-name')

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_normal_address_rename(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = ['old-name']

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 0)

        address_book.rename.assert_called_once_with('old-name', 'new-name')

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_doesnt_exits(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 1)

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_duplicate_destination_key_name(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['new-name']

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 1)

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_duplicate_destination_address_name(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = ['new-name']

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 1)

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_corrupted_db_duplicate_address_key(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['old-name']

        address_book = MockAddressBook()
        address_book.keys.return_value = ['old-name']

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        with self.assertRaises(RuntimeError):
            run_rename(args)

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_handle_address_rename_failure(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = ['old-name']
        address_book.rename.return_value = False

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 1)

        address_book.rename.assert_called_once_with('old-name', 'new-name')

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_handle_key_rename_failure(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['old-name']
        key_store.rename_key.return_value = False

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.old = 'old-name'
        args.new = 'new-name'

        from pocketbook.commands.rename import run_rename
        self.assertEqual(run_rename(args), 1)

        key_store.rename_key.assert_called_once_with('old-name', 'new-name')



import os
import shutil
import tempfile

SAMPLE_ADDRESS = '2w9uqXLFvCve9TjfoRb2nuztuTx54E1S3AjuB9hWCMMGX5kp4W'
SUPER_SECURE_PASSWORD = 'Fetch!Ai-ToTh3M00n!'


class TemporaryPocketBookRoot:
    def __init__(self):
        self._root = None

    @property
    def root(self):
        return self._root

    def __enter__(self):
        self._root = tempfile.mkdtemp(prefix='pocketbook_', suffix='_root')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if os.path.isdir(self._root):
            shutil.rmtree(self._root)



import unittest
from unittest.mock import patch, Mock, MagicMock, PropertyMock

from fetchai.ledger.crypto import Entity, Address
from fetchai.ledger.serialisation.transaction import encode_transaction

from pocketbook.address_book import AddressBook
from pocketbook.key_store import KeyStore
from pocketbook.utils import create_api


class Person:
    def __init__(self, name):
        self.name = str(name)
        self.entity = Entity()
        self.address = Address(self.entity)


class TransferCommandUnitTests(unittest.TestCase):

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_transfer_to_addr_dest(self, MockKeyStore, MockAddressBook, mock_create_api, mock_encode_transaction,
                                   *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name]
        key_store.load_key.return_value = person1.entity

        address_book = MockAddressBook()
        address_book.keys.return_value = [person2.name]
        address_book.lookup_address.return_value = person2.address

        api = MagicMock()
        mock_create_api.return_value = api

        tx = MagicMock()
        tx.charge_rate = PropertyMock()
        api.tokens._create_skeleton_tx.return_value = tx
        api.tokens._post_tx_json.return_value = '0xTransactionHexId'

        encoded_tx = MagicMock()
        mock_encode_transaction.return_value = encoded_tx

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = None
        args.network = 'super-duper-net'

        from pocketbook.commands.transfer import run_transfer
        run_transfer(args)

        mock_create_api.assert_called_once_with('super-duper-net')
        key_store.load_key.assert_called_once_with(person1.name, 'weak-password')
        api.tokens._create_skeleton_tx.assert_called_once_with(1)  # single signature
        tx.add_transfer.assert_called_once_with(person2.address, 20000000000)
        mock_encode_transaction.assert_called_once_with(tx, [person1.entity])
        api.tokens._post_tx_json.assert_called_once_with(encoded_tx, 'transfer')
        api.sync.assert_called_once_with('0xTransactionHexId')

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_transfer_to_key_dest(self, MockKeyStore, MockAddressBook, mock_create_api, mock_encode_transaction, *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name, person2.name]
        key_store.load_key.side_effect = [person1.entity, person2.entity]
        key_store.lookup_address.return_value = person2.address

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        api = MagicMock()
        mock_create_api.return_value = api

        tx = MagicMock()
        api.tokens._create_skeleton_tx.return_value = tx
        api.tokens._post_tx_json.return_value = '0xTransactionHexId'

        encoded_tx = MagicMock()
        mock_encode_transaction.return_value = encoded_tx

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = None
        args.network = 'super-duper-net'

        from pocketbook.commands.transfer import run_transfer
        run_transfer(args)

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_transfer_to_new_dest(self, MockKeyStore, MockAddressBook, mock_create_api, mock_encode_transaction, *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name, person2.name]
        key_store.load_key.side_effect = [person1.entity]
        key_store.lookup_address.return_value = None

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        api = MagicMock()
        mock_create_api.return_value = api

        tx = MagicMock()
        api.tokens._create_skeleton_tx.return_value = tx
        api.tokens._post_tx_json.return_value = '0xTransactionHexId'

        encoded_tx = MagicMock()
        mock_encode_transaction.return_value = encoded_tx

        args = Mock()
        args.destination = str(person2.address)
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = None
        args.network = 'super-duper-net'

        from pocketbook.commands.transfer import run_transfer
        run_transfer(args)

        mock_create_api.assert_called_once_with('super-duper-net')
        key_store.load_key.assert_called_once_with(person1.name, 'weak-password')
        api.tokens._create_skeleton_tx.assert_called_once_with(1)  # single signature
        tx.add_transfer.assert_called_once_with(person2.address, 20000000000)
        mock_encode_transaction.assert_called_once_with(tx, [person1.entity])
        api.tokens._post_tx_json.assert_called_once_with(encoded_tx, 'transfer')
        api.sync.assert_called_once_with('0xTransactionHexId')

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_multisig_transfer(self, MockKeyStore, MockAddressBook, mock_create_api, mock_encode_transaction, *args):
        multisig = Person('MultiSig')
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name, person2.name]
        key_store.load_key.side_effect = [person1.entity]
        key_store.lookup_address.return_value = person2.address

        address_book = MockAddressBook()
        address_book.keys.return_value = [multisig.name]
        address_book.lookup_address.return_value = multisig.address

        api = MagicMock()
        mock_create_api.return_value = api

        tx = MagicMock()
        api.tokens._create_skeleton_tx.return_value = tx
        api.tokens._post_tx_json.return_value = '0xTransactionHexId'

        encoded_tx = MagicMock()
        mock_encode_transaction.return_value = encoded_tx

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = multisig.name
        args.network = 'super-duper-net'

        from pocketbook.commands.transfer import run_transfer
        run_transfer(args)

        mock_create_api.assert_called_once_with('super-duper-net')
        address_book.lookup_address.assert_called_once_with(multisig.name)
        key_store.lookup_address.assert_called_once_with(person2.name)
        key_store.load_key.assert_called_once_with(person1.name, 'weak-password')
        api.tokens._create_skeleton_tx.assert_called_once_with(1)  # single signature
        tx.add_transfer.assert_called_once_with(person2.address, 20000000000)
        mock_encode_transaction.assert_called_once_with(tx, [person1.entity])
        api.tokens._post_tx_json.assert_called_once_with(encoded_tx, 'transfer')
        api.sync.assert_called_once_with('0xTransactionHexId')

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_error_when_signer_not_present(self, MockKeyStore, MockAddressBook, mock_create_api,
                                           mock_encode_transaction, *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = []
        key_store.load_key.side_effect = [person1.entity]
        key_store.lookup_address.return_value = person2.address

        address_book = MockAddressBook()
        address_book.keys.return_value = [person1.name, person2.name]
        address_book.lookup_address.return_value = person2.address

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = person2.name
        args.network = 'super-duper-net'

        with self.assertRaises(RuntimeError):
            from pocketbook.commands.transfer import run_transfer
            run_transfer(args)

        # mock_create_api.assert_called_once_with('super-duper-net')
        address_book.lookup_address.assert_called_once_with(person2.name)

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_error_when_from_field_is_invalid(self, MockKeyStore, MockAddressBook, mock_create_api,
                                              mock_encode_transaction, *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name]
        key_store.load_key.side_effect = [person1.entity]
        key_store.lookup_address.return_value = person2.address

        address_book = MockAddressBook()
        address_book.keys.return_value = [person2.name]
        address_book.lookup_address.return_value = person2.address

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = [person1.name]
        args.from_address = 'some-one-missing'
        args.network = 'super-duper-net'

        with self.assertRaises(RuntimeError):
            from pocketbook.commands.transfer import run_transfer
            run_transfer(args)

        # mock_create_api.assert_called_once_with('super-duper-net')
        address_book.lookup_address.assert_called_once_with(person2.name)

    @patch('getpass.getpass', side_effect=['weak-password'])
    @patch('builtins.input', return_value='')
    @patch('fetchai.ledger.serialisation.transaction.encode_transaction', spec=encode_transaction)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_error_case_with_bad_args(self, MockKeyStore, MockAddressBook, mock_create_api, mock_encode_transaction,
                                      *args):
        person1 = Person('Jane')
        person2 = Person('Clare')

        key_store = MockKeyStore()
        key_store.list_keys.return_value = [person1.name]
        key_store.load_key.side_effect = [person1.entity]
        key_store.lookup_address.return_value = person2.address

        address_book = MockAddressBook()
        address_book.keys.return_value = [person2.name]
        address_book.lookup_address.return_value = person2.address

        args = Mock()
        args.destination = person2.name
        args.amount = 20000000000
        args.charge_rate = 1
        args.signers = []
        args.from_address = 'some-one-missing'
        args.network = 'super-duper-net'

        with self.assertRaises(RuntimeError):
            from pocketbook.commands.transfer import run_transfer
            run_transfer(args)

        # mock_create_api.assert_called_once_with('super-duper-net')
        address_book.lookup_address.assert_called_once_with(person2.name)



import os
from copy import deepcopy

import toml
from fetchai.ledger.crypto import Entity, Address

from .constants import DEFAULT_KEY_STORE_ROOT


class KeyStoreError(Exception):
    pass


class DuplicateKeyNameError(KeyStoreError):
    def __str__(self):
        return 'Duplicate key name found'


class KeyNotFoundError(KeyStoreError):
    def __str__(self):
        return 'Unable to find key in store'


class UnableToDecodeKeyError(KeyStoreError):
    def __str__(self):
        return 'Unable to decode key from key store'


class KeyStore:
    INDEX_FILE_NAME = 'index.toml'

    def __init__(self, root=None):
        self._root = root or DEFAULT_KEY_STORE_ROOT
        self._index_path = os.path.join(self._root, self.INDEX_FILE_NAME)

        os.makedirs(self._root, exist_ok=True)

        if os.path.exists(self._index_path):
            with open(self._index_path, 'r') as index_file:
                self._index = toml.load(index_file)
        else:
            self._index = {'key': []}

    def list_keys(self):
        return list([item['name'] for item in self._index['key']])

    def lookup_address(self, name):
        address = None
        for item in self._index['key']:
            if item['name'] == name:
                address = item['address']
                break
        return address

    def load_key(self, name: str, password: str) -> Entity:
        if name not in self.list_keys():
            raise KeyNotFoundError()

        key_file_path = self._format_key_path(name)
        assert os.path.exists(key_file_path)

        with open(key_file_path, 'r') as key_file:
            entity = Entity.load(key_file, password)
            address = Address(entity)

            # lookup the metadata for this
            metadata = self._lookup_meta_data(name)
            if metadata['address'] != str(address):
                raise UnableToDecodeKeyError()

            return entity

    def add_key(self, name: str, password: str, entity: Entity):
        if name in self.list_keys():
            raise DuplicateKeyNameError()

        key_file_path = self._format_key_path(name)
        assert not os.path.exists(key_file_path)

        # dump out the file
        with open(key_file_path, 'w') as key_file:
            entity.dump(key_file, password)

        # update the index
        self._index['key'].append({
            'name': name,
            'address': str(Address(entity)),
        })
        self._flush_index()

    def rename_key(self, old_name: str, new_name: str) -> bool:

        # do some basic checks
        if old_name not in self.list_keys():
            return False
        if new_name in self.list_keys():
            return False

        # extract the old key metadata
        old_metadata = self._lookup_meta_data(old_name)
        if old_metadata is None:
            return False

        # move the key file
        old_key_file_path = self._format_key_path(old_name)
        new_key_file_path = self._format_key_path(new_name)
        os.rename(old_key_file_path, new_key_file_path)

        # remove the old entries to the metadata
        self._remove_key_from_index(old_name)

        # update and insert the new meta data into place
        new_metadata = deepcopy(old_metadata)
        new_metadata['name'] = new_name
        self._index['key'].append(new_metadata)
        self._flush_index()

        return True

    def remove_key(self, name: str) -> bool:
        if name not in self.list_keys():
            return False

        # remove the key path from the disk
        key_file_path = self._format_key_path(name)
        if not os.path.exists(key_file_path):
            raise RuntimeError('Key not present on disk, unable to remove')
        os.remove(key_file_path)

        # update the index
        self._remove_key_from_index(name)
        self._flush_index()

        return True

    def _remove_key_from_index(self, name: str):
        self._index['key'] = list(filter(lambda x: x['name'] != name, self._index['key']))

    def _format_key_path(self, name: str):
        return os.path.join(self._root, '{}.key'.format(name))

    def _lookup_meta_data(self, name: str):
        for item in self._index['key']:
            if item['name'] == name:
                return item

    def _flush_index(self):
        with open(self._index_path, 'w') as index_file:
            toml.dump(self._index, index_file)



import argparse
import sys

from . import __version__
from .commands.add import run_add
from .commands.create import run_create
from .commands.delete import run_delete
from .commands.list import run_list
from .commands.rename import run_rename
from .commands.transfer import run_transfer
from .disclaimer import display_disclaimer
from .utils import NetworkUnavailableError, checked_address, to_canonical, MINIMUM_FRACTIONAL_FET


def parse_commandline():
    parser = argparse.ArgumentParser(prog='pocketbook')
    parser.add_argument('-v', '--version', action='version', version=__version__)
    parser.add_argument('-n', '--network', default='mainnet', help='The name of the target being addressed')
    subparsers = parser.add_subparsers()

    parser_list = subparsers.add_parser('list', aliases=['ls'], help='Lists all the balances and addresses')
    parser_list.add_argument('-v', '--verbose', action='store_true', help='Display extra information (if available)')
    parser_list.add_argument('pattern', nargs='*', default=['*'])
    parser_list.set_defaults(handler=run_list)

    parser_create = subparsers.add_parser('create', aliases=['new'], help='Create a new key')
    parser_create.set_defaults(handler=run_create)

    parser_add = subparsers.add_parser('add', help='Adds an address to the address book')
    parser_add.add_argument('name', help='The name of the key')
    parser_add.add_argument('address', type=checked_address, help='The account address')
    parser_add.set_defaults(handler=run_add)

    parser_transfer = subparsers.add_parser('transfer', help='Transfers funds between wallets')
    parser_transfer.add_argument('destination',
                                 help='The destination address either a name in the address book or an address')
    parser_transfer.add_argument('amount', type=to_canonical, help='The amount of FET to be transferred')
    parser_transfer.add_argument('--from', dest='from_address', help='The signing account, required for multi-sig')
    parser_transfer.add_argument('-R', '--charge-rate', type=to_canonical, default=to_canonical(MINIMUM_FRACTIONAL_FET),
                                 help='The charge rate associated with this transaction')
    parser_transfer.add_argument('signers', nargs='+', help='The series of key names needed to sign the transaction')
    parser_transfer.set_defaults(handler=run_transfer)

    parser_rename = subparsers.add_parser('rename', aliases=['mv'], help='Renames and address or key to another name')
    parser_rename.add_argument('old', help='The name of the old account name')
    parser_rename.add_argument('new', help='The new name of the account')
    parser_rename.set_defaults(handler=run_rename)

    parser_delete = subparsers.add_parser('delete', aliases=['rm'], help='Deletes an address or key from the wallet')
    parser_delete.add_argument('name', help='The name of the account to remove')
    parser_delete.set_defaults(handler=run_delete)

    return parser, parser.parse_args()


def main():
    # run the specified command
    exit_code = 1
    try:
        display_disclaimer()

        # parse the command line
        parser, args = parse_commandline()

        # select the command handler
        if hasattr(args, 'handler'):
            handler = args.handler
        else:
            parser.print_usage()
            handler = None

        if handler is not None:
            # execute the handler
            ret = handler(args)

            # update the exit code based on the return value
            if isinstance(ret, int):
                exit_code = ret
            else:
                exit_code = 0

    except KeyboardInterrupt:
        print('Stopped!')

    except NetworkUnavailableError:
        print('The network appears to be unavailable at the moment. Please try again later')

    except Exception as ex:
        print('Error:', ex)

    # close the program with the given exit code
    sys.exit(int(exit_code))



from typing import Union

from fetchai.ledger.crypto import Address

CANONICAL_FET_UNIT = 1e10
MINIMUM_FRACTIONAL_FET = 1 / CANONICAL_FET_UNIT


class NetworkUnavailableError(Exception):
    pass


class ConversionError(RuntimeError):
    def __init__(self, msg):
        super().__init__(msg)


def to_canonical(value: Union[float, int]) -> int:
    value = float(value)
    if value < 0:
        raise ConversionError('Unable to convert negative token amount: {}'.format(value))
    if value < MINIMUM_FRACTIONAL_FET:
        raise ConversionError(
            'Converted value {} is below minimum transfer value: {}'.format(value, MINIMUM_FRACTIONAL_FET))

    return int(value * CANONICAL_FET_UNIT)


def from_canonical(value: int) -> float:
    value = int(value)
    if value < 0:
        raise ConversionError('Unable to convert negative token amount: {}'.format(value))

    return value / int(CANONICAL_FET_UNIT)


def token_amount(value: float) -> str:
    """
    Converts a token amount into a fixed precision string value
    :param value: The input value to display
    :return: The converted value
    """
    return '{:6.10f} FET'.format(float(value))


def get_balance(api, address):
    balance = int(api.tokens.balance(address))
    return from_canonical(balance)


def get_stake(api, addresss):
    stake = int(api.tokens.stake(addresss))
    return from_canonical(stake)


def create_api(name: str):
    from fetchai.ledger.api import LedgerApi

    try:
        return LedgerApi(network=name)
    except:
        pass

    raise NetworkUnavailableError()


def checked_address(address):
    try:
        return Address(address)
    except:
        raise RuntimeError(
            'Unable to convert {} into and address. The address needs to be a base58 encoded value'.format(address))



from colored import fg, bg, attr


class Table:
    def __init__(self, columns):
        self._rows = []
        self._columns = columns

    def add_row(self, **kwargs):
        self._rows.append(dict(**kwargs))

    def display(self, column_padding=1):

        # Step 1. Determine the column lengths
        column_lengths = [0] * len(self._columns)
        for i, column in enumerate(self._columns):

            # header
            column_lengths[i] = len(column)

            # row contents
            for row in self._rows:
                column_lengths[i] = max(column_lengths[i], len(str(row.get(column, ''))))

            # allow a nice
            column_lengths[i] += column_padding

        # Step 2a. Print the header
        self._start_header()
        for i, column in enumerate(self._columns):
            self._write_header(column, column_lengths[i])
        self._end_header()

        # Step 2b. Print the table contents
        for i, row in enumerate(self._rows):
            self._start_row(i)
            for j, column in enumerate(self._columns):
                self._write_row_value(row.get(column, ''), i, column_lengths[j])
            self._end_row(i)

    @classmethod
    def _start_header(cls):
        print(fg(255), end='')

    @classmethod
    def _write_header(cls, value, length):
        value = cls._pad_value(value, length)

        # capitalise
        value = value[0].upper() + value[1:]
        print(value, end='')

    @classmethod
    def _end_header(cls):
        print(attr(0))

    @classmethod
    def _start_row(cls, row):
        if row & 1:
            foreground = fg(251)
            background = bg(235)
        else:
            foreground = fg(253)
            background = bg(238)

        print(foreground + background, end='')

    @classmethod
    def _write_row_value(cls, value, row, length):
        value = cls._pad_value(value, length)

        print(value, end='')

    @classmethod
    def _end_row(cls, row):
        print(attr(0))

    @classmethod
    def _pad_value(cls, value, length):
        value = str(value)
        in_value = str(value)
        value_len = len(value)

        if value_len > length:
            value = value[:length]
        else:
            padding_len = length - value_len
            value += ' ' * padding_len

        return value



import unittest
from unittest.mock import patch, Mock

from pocketbook.address_book import AddressBook
from pocketbook.key_store import KeyStore
from tests.utils import SAMPLE_ADDRESS


class DeleteCommandTests(unittest.TestCase):

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_address(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = ['foo']

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(0, run_delete(args))

        address_book.remove.assert_called_once_with('foo')

    @patch('builtins.input', return_value='foo-address')
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_key(self, MockKeyStore, MockAddressBook, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['foo']
        key_store.lookup_address.return_value = 'foo-address'

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(0, run_delete(args))

        key_store.lookup_address.assert_called_once_with('foo')
        key_store.remove_key.assert_called_once_with('foo')

    @patch('builtins.input', return_value='foo-address')
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_unknown(self, MockKeyStore, MockAddressBook, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(1, run_delete(args))

    @patch('builtins.input', return_value='not-sample-address')
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_key_invalid_confirmation(self, MockKeyStore, MockAddressBook, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['foo']
        key_store.lookup_address.return_value = SAMPLE_ADDRESS

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(1, run_delete(args))

        key_store.lookup_address.assert_called_once_with('foo')

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_address_failed(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        address_book = MockAddressBook()
        address_book.keys.return_value = ['foo']
        address_book.remove.return_value = False

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(1, run_delete(args))

        address_book.remove.assert_called_once_with('foo')

    @patch('builtins.input', return_value='foo-address')
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_delete_key_failed(self, MockKeyStore, MockAddressBook, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['foo']
        key_store.lookup_address.return_value = 'foo-address'
        key_store.remove_key.return_value = False

        address_book = MockAddressBook()
        address_book.keys.return_value = []

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        self.assertEqual(1, run_delete(args))

        key_store.lookup_address.assert_called_once_with('foo')
        key_store.remove_key.assert_called_once_with('foo')

    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_db_corruption(self, MockKeyStore, MockAddressBook):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['foo']

        address_book = MockAddressBook()
        address_book.keys.return_value = ['foo']

        args = Mock()
        args.name = 'foo'

        from pocketbook.commands.delete import run_delete
        with self.assertRaises(RuntimeError):
            run_delete(args)



import unittest
from typing import Union
from unittest.mock import MagicMock, patch

from fetchai.ledger.crypto import Address, Entity

from pocketbook.utils import get_balance, get_stake, NetworkUnavailableError, checked_address, to_canonical, \
    from_canonical, token_amount, ConversionError


class UtilsTests(unittest.TestCase):

    def test_get_balance(self):
        api = MagicMock()
        api.tokens.balance.return_value = to_canonical(10)

        balance = get_balance(api, 'some address')
        self.assertEqual(balance, 10)

        api.tokens.balance.assert_called_once_with('some address')

    def test_get_stake(self):
        api = MagicMock()
        api.tokens.stake.return_value = to_canonical(5)

        stake = get_stake(api, 'some address')
        self.assertEqual(stake, 5)

        api.tokens.stake.assert_called_once_with('some address')

    @patch('fetchai.ledger.api.LedgerApi')
    def test_create_api(self, MockLedgerApi):
        # normal use
        from pocketbook.utils import create_api
        create_api('super-duper-net')

        MockLedgerApi.assert_called_once_with(network='super-duper-net')
        MockLedgerApi.reset_mock()

    @patch('fetchai.ledger.api.LedgerApi', side_effect=[RuntimeError('Bad Error')])
    def test_error_on_create_api(self, MockLedgerApi):
        # internal error case
        from pocketbook.utils import create_api
        with self.assertRaises(NetworkUnavailableError):
            create_api('super-duper-net')

        MockLedgerApi.assert_called_once_with(network='super-duper-net')

    def test_valid_address(self):
        entity = Entity()
        address = Address(entity)

        recovered_address = checked_address(str(address))

        self.assertEqual(str(recovered_address), str(address))

    def test_invalid_address_exception(self):
        with self.assertRaises(RuntimeError) as ctx:
            checked_address('foo-bar-baz')
        self.assertEqual(str(ctx.exception),
                         'Unable to convert foo-bar-baz into and address. The address needs to be a base58 encoded value')


class FetConversionTests(unittest.TestCase):
    def assertIsConvertible(self, canonical: int, value: Union[int, float]):
        converted_canonical = to_canonical(value)
        self.assertEqual(canonical, converted_canonical)
        self.assertEqual(float(value), from_canonical(converted_canonical))

    def test_canonical_conversions(self):
        self.assertIsConvertible(10000000000, 1)
        self.assertIsConvertible(10000000000, 1.0)
        self.assertIsConvertible(12000000000, 1.2)
        self.assertIsConvertible(10020000000, 1.002)
        self.assertIsConvertible(1, 1e-10)
        self.assertIsConvertible(100, 1e-8)
        self.assertIsConvertible(10000000, 1e-3)
        self.assertIsConvertible(10000, 1e-6)
        self.assertIsConvertible(10, 1e-9)
        self.assertIsConvertible(10000000000000, 1e3)
        self.assertIsConvertible(10000000000000000, 1e6)
        self.assertIsConvertible(10000000000000000000, 1e9)

    def test_token_amount_formatting(self):
        self.assertEqual(token_amount(1), '1.0000000000 FET')
        self.assertEqual(token_amount(1.0), '1.0000000000 FET')
        self.assertEqual(token_amount(1.2), '1.2000000000 FET')
        self.assertEqual(token_amount(1.002), '1.0020000000 FET')
        self.assertEqual(token_amount(1e-10), '0.0000000001 FET')
        self.assertEqual(token_amount(1e-8), '0.0000000100 FET')
        self.assertEqual(token_amount(1e-3), '0.0010000000 FET')
        self.assertEqual(token_amount(1e-6), '0.0000010000 FET')
        self.assertEqual(token_amount(1e-9), '0.0000000010 FET')
        self.assertEqual(token_amount(1e3), '1000.0000000000 FET')
        self.assertEqual(token_amount(1e6), '1000000.0000000000 FET')
        self.assertEqual(token_amount(1e9), '1000000000.0000000000 FET')

    def test_invalid_negative_to_canonical(self):
        with self.assertRaises(ConversionError):
            to_canonical(-10)

    def test_invalid_too_small_to_canonical(self):
        with self.assertRaises(ConversionError):
            to_canonical(1e-20)

    def test_invalid_non_number_to_canonical(self):
        with self.assertRaises(ValueError):
            to_canonical('foo-bar')

    def test_invalid_negative_from_canonical(self):
        with self.assertRaises(ConversionError):
            from_canonical(-10)



import os

import toml

from .constants import DEFAULT_KEY_STORE_ROOT


class AddressBook:
    INDEX_FILE_NAME = 'addresses.toml'

    def __init__(self, root=None):
        self._root = root or DEFAULT_KEY_STORE_ROOT
        self._address_book_path = os.path.join(self._root, self.INDEX_FILE_NAME)
        self._address_book = self._load()

    def add(self, name, address):
        if name in self._address_book:
            raise RuntimeError('Address already exists')

        self._address_book[name] = str(address)
        self._save()

    def rename(self, old: str, new: str) -> bool:
        """
        Rename an address from one name to another

        :param old: The old name for the address
        :param new: The new name for the address
        :return: True if successful, otherwise False
        """
        success = False

        if old in self._address_book and new not in self._address_book:
            self._address_book[new] = self._address_book[old]
            del self._address_book[old]
            self._save()
            success = True

        return success

    def remove(self, name: str) -> bool:
        """
        Remove an address from the address book

        :param name: The name of the address to remove
        :return: True if successful, otherwise False
        """
        success = False
        if name in self._address_book:
            del self._address_book[name]
            self._save()
            success = True
        return success

    def keys(self):
        return self._address_book.keys()

    def items(self):
        return self._address_book.items()

    def lookup_address(self, name):
        if name not in self._address_book:
            raise RuntimeError('Unable to lookup requested address: {}'.format(name))

        return self._address_book[name]

    def _save(self):
        with open(self._address_book_path, 'w') as output_file:
            toml.dump(self._address_book, output_file)

    def _load(self):
        if not os.path.exists(self._address_book_path):
            return {}
        with open(self._address_book_path, 'r') as input_file:
            return toml.load(input_file)



import os
import unittest
from unittest.mock import patch

import toml
from fetchai.ledger.crypto import Entity, Address

from pocketbook.key_store import KeyStore, DuplicateKeyNameError, KeyNotFoundError, UnableToDecodeKeyError
from .utils import TemporaryPocketBookRoot, SUPER_SECURE_PASSWORD


class KeyStoreTestCase(unittest.TestCase):
    def assertKeyIsNotPresentOnDisk(self, name: str, ctx: TemporaryPocketBookRoot):
        self.assertTrue(os.path.isdir(ctx.root))

        # check the index record
        index_path = os.path.join(ctx.root, KeyStore.INDEX_FILE_NAME)
        self.assertTrue(os.path.isfile(index_path))
        with open(index_path, 'r') as index_file:
            index = toml.load(index_path)

            # build up the key listings
            key_addresses = {key['name']: key['address'] for key in index.get('key', [])}
            self.assertNotIn(name, key_addresses)

        key_path = os.path.join(ctx.root, '{}.key'.format(name))
        self.assertFalse(os.path.isfile(key_path))

    def assertKeyIsPresentOnDisk(self, name: str, password: str, entity: Entity, ctx: TemporaryPocketBookRoot):
        self.assertTrue(os.path.isdir(ctx.root))

        # check that the index record has been created correctly
        index_path = os.path.join(ctx.root, KeyStore.INDEX_FILE_NAME)
        self.assertTrue(os.path.isfile(index_path))
        with open(index_path, 'r') as index_file:
            index = toml.load(index_path)

            # build up the key listings
            key_addresses = {key['name']: key['address'] for key in index.get('key', [])}

            # ensure the key is reference in the index
            self.assertIn(name, key_addresses.keys())

            # also ensure the computed address is also correct
            expected_address = Address(entity)
            self.assertEqual(key_addresses[name], str(expected_address))

        # check the contents of the key
        key_path = os.path.join(ctx.root, '{}.key'.format(name))
        self.assertTrue(os.path.isfile(key_path))
        with open(key_path, 'r') as key_file:
            recovered = Entity.load(key_file, password)
            self.assertEqual(recovered.private_key, entity.private_key)


class KeyStoreTests(KeyStoreTestCase):

    def test_init(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)
            self.assertEqual(len(key_store.list_keys()), 0)

    def test_add_key(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)

            # create an entity and add it to the store
            entity = Entity()
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            self.assertEqual(len(key_store.list_keys()), 1)

    def test_added_key_is_flushed_to_disk(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)

            # create an entity and add it to the store
            entity = Entity()
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            # check the files on have been set correctly
            self.assertKeyIsPresentOnDisk('sample', SUPER_SECURE_PASSWORD, entity, ctx)

    def test_address_lookup(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            expected_address = Address(entity)
            recovered_address = key_store.lookup_address('sample')

            self.assertEqual(str(expected_address), str(recovered_address))

    def test_duplicate_adds(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            with self.assertRaises(DuplicateKeyNameError):
                key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

    def test_loading_of_multiple_stores(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store1 = KeyStore(root=ctx.root)
            key_store1.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            key_store2 = KeyStore(root=ctx.root)
            self.assertEqual(set(key_store2.list_keys()), set(key_store1.list_keys()))

    def test_loading_saved_key(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            recovered = key_store.load_key('sample', SUPER_SECURE_PASSWORD)
            self.assertEqual(recovered.private_key, entity.private_key)

    def test_loading_invalid_key(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)

            with self.assertRaises(KeyNotFoundError):
                key_store.load_key('foo-bar', SUPER_SECURE_PASSWORD)

    def test_unable_to_load_saved_key(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            with self.assertRaises(UnableToDecodeKeyError):
                key_store.load_key('sample', '!' + SUPER_SECURE_PASSWORD)

    def test_duplicate_key_error_message(self):
        error = DuplicateKeyNameError()
        self.assertEqual(str(error), 'Duplicate key name found')

    def test_key_not_found_error_message(self):
        error = KeyNotFoundError()
        self.assertEqual(str(error), 'Unable to find key in store')

    def test_unable_to_decode_key_error_message(self):
        error = UnableToDecodeKeyError()
        self.assertEqual(str(error), 'Unable to decode key from key store')

    def test_rename(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            self.assertTrue(key_store.rename_key('sample', 'sample2'))
            self.assertIn('sample2', key_store.list_keys())
            self.assertNotIn('sample', key_store.list_keys())

    def test_rename_is_persisted_to_disk(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()
            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            self.assertTrue(key_store.rename_key('sample', 'sample2'))
            self.assertKeyIsPresentOnDisk('sample2', SUPER_SECURE_PASSWORD, entity, ctx)

    def test_rename_failure_doesnt_exist(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)
            self.assertFalse(key_store.rename_key('sample', 'sample2'))

    def test_rename_failure_already_exists(self):
        with TemporaryPocketBookRoot() as ctx:
            entity1 = Entity()
            entity2 = Entity()

            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample1', SUPER_SECURE_PASSWORD, entity1)
            key_store.add_key('sample2', SUPER_SECURE_PASSWORD, entity2)

            self.assertFalse(key_store.rename_key('sample1', 'sample2'))

    def test_corrupted_metadata(self):
        with TemporaryPocketBookRoot() as ctx:
            entity1 = Entity()

            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample1', SUPER_SECURE_PASSWORD, entity1)

            with patch.object(key_store, '_lookup_meta_data', return_value=None):
                self.assertFalse(key_store.rename_key('sample1', 'sample2'))

    def test_remove_failed(self):
        with TemporaryPocketBookRoot() as ctx:
            key_store = KeyStore(root=ctx.root)
            self.assertFalse(key_store.remove_key('sample'))

    def test_remove(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()

            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            self.assertTrue(key_store.remove_key('sample'))
            self.assertNotIn('sample', key_store.list_keys())

    def test_remove_reflected_on_disk(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()

            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            self.assertTrue(key_store.remove_key('sample'))
            self.assertKeyIsNotPresentOnDisk('sample', ctx)

    def test_remove_with_corrupted_files(self):
        with TemporaryPocketBookRoot() as ctx:
            entity = Entity()

            key_store = KeyStore(root=ctx.root)
            key_store.add_key('sample', SUPER_SECURE_PASSWORD, entity)

            # corrupt the files
            key_path = key_store._format_key_path('sample')
            self.assertTrue(os.path.isfile(key_path))
            os.remove(key_path)

            with self.assertRaises(RuntimeError):
                key_store.remove_key('sample')



import unittest
from io import StringIO
from unittest.mock import patch

from pocketbook.table import Table


class TableTests(unittest.TestCase):

    @patch('sys.stdout', new_callable=StringIO)
    def test_empty(self, mock_print):
        table = Table(['foo'])
        table.display()

        self.assertEqual(mock_print.getvalue(), '\x1b[38;5;255mFoo \x1b[0m\n')

    @patch('sys.stdout', new_callable=StringIO)
    def test_columns(self, mock_print):
        table = Table(['foo', 'bar'])
        table.add_row(foo=1, bar=2)
        table.display()

        self.assertEqual(mock_print.getvalue(),
                         '\x1b[38;5;255mFoo Bar \x1b[0m\n\x1b[38;5;253m\x1b[48;5;238m1   2   \x1b[0m\n')

    @patch('sys.stdout', new_callable=StringIO)
    def test_multiple_rows(self, mock_print):
        table = Table(['foo'])
        table.add_row(foo=1)
        table.add_row(foo=2)
        table.add_row(foo=3)
        table.display()

        self.assertEqual(mock_print.getvalue(),
                         '\x1b[38;5;255mFoo \x1b[0m\n\x1b[38;5;253m\x1b[48;5;238m1   \x1b[0m\n\x1b[38;5;251m\x1b[48;5;235m2   \x1b[0m\n\x1b[38;5;253m\x1b[48;5;238m3   \x1b[0m\n')

    def test_invalid_padding_case(self):
        truncated = Table._pad_value('LongWords', 5)
        self.assertEqual(truncated, 'LongW')



def run_create(args):
    from getpass import getpass
    from fetchai.ledger.crypto.entity import _strong_password as is_strong_password
    from fetchai.ledger.crypto import Entity
    from pocketbook.key_store import KeyStore

    key_store = KeyStore()
    existing_keys = set(key_store.list_keys())

    # get the name for the new key
    while True:
        name = input('Enter name for key: ')
        if name in existing_keys:
            print('Key name already exists')
            continue
        break

    # prompt the user for the password
    while True:
        password = getpass('Enter password for key...: ')
        if not is_strong_password(password):
            print('Password too simple, try again')
            continue

        confirm = getpass('Confirm password for key.: ')
        if password != confirm:
            print('Passwords did not match, try again')
            continue

        break

    key_store.add_key(name, password, Entity())



from fnmatch import fnmatch


def _should_display(item, patterns):
    return any([fnmatch(item, p) for p in patterns])


def run_list(args):
    from pocketbook.address_book import AddressBook
    from pocketbook.key_store import KeyStore
    from pocketbook.table import Table
    from pocketbook.utils import create_api, get_balance, get_stake

    address_book = AddressBook()
    key_store = KeyStore()
    keys = key_store.list_keys()

    if len(keys) == 0:
        print('No keys present')
    else:

        # select the columns
        cols = ['name', 'type', 'balance', 'stake']
        if args.verbose:
            cols.append('address')

        api = create_api(args.network)

        table = Table(cols)
        for key in keys:
            if not _should_display(key, args.pattern):
                continue

            address = key_store.lookup_address(key)
            balance = get_balance(api, address)
            stake = get_stake(api, address)

            row_data = {
                'name': key,
                'type': 'key',
                'balance': balance,
                'stake': stake,
                'address': str(address),
            }

            table.add_row(**row_data)

        for name, address in address_book.items():
            if not _should_display(name, args.pattern):
                continue

            balance = get_balance(api, address)
            stake = get_stake(api, address)

            row_data = {
                'name': name,
                'type': 'addr',
                'balance': balance,
                'stake': stake,
                'address': str(address),
            }

            table.add_row(**row_data)

        table.display()



KEY_WARNING_TEMPLATE = """
### WARNING ###

You are about to delete the key {} from the database. This is a PERMANENT change and can
no be reverted. This might mean that you loose access to funds forever.

If you want to continue please confirm by entering the address of wallet you want to remove
below:

Address (Base58): """


def run_delete(args):
    from pocketbook.key_store import KeyStore
    from pocketbook.address_book import AddressBook

    key_store = KeyStore()
    address_book = AddressBook()

    is_key = args.name in key_store.list_keys()
    is_address = args.name in address_book.keys()

    def address_delete():
        return address_book.remove(args.name)

    def key_delete():
        return key_store.remove_key(args.name)

    if is_key and is_address:
        raise RuntimeError('Corrected database, account is both an address and a key')
    elif is_key:
        warning = True
        handler = key_delete
    elif is_address:
        warning = False
        handler = address_delete
    else:
        print('Unknown key or address: {}. Please check and try again'.format(args.name))
        return 1

    # double check that the user really want to remove the key
    if warning:
        input_address = input(KEY_WARNING_TEMPLATE.format(args.name))

        address = key_store.lookup_address(args.name)

        if str(address) != input_address:
            print('The input address for key {} does not match. Please double check and try again')
            return 1

    # perform the removal
    if not handler():
        print('Failed to remove the specified key or address')
        return 1

    return 0



import os
import unittest

import toml

from pocketbook.address_book import AddressBook
from .utils import TemporaryPocketBookRoot, SAMPLE_ADDRESS


class AddressBookTests(unittest.TestCase):
    def assertAddressIsNotPresentOnDisk(self, name: str, ctx: TemporaryPocketBookRoot):
        address_index_path = os.path.join(ctx.root, AddressBook.INDEX_FILE_NAME)
        if os.path.isfile(address_index_path):
            with open(address_index_path, 'r') as address_index_file:
                address_index = toml.load(address_index_file)
                self.assertNotIn(name, address_index)

    def assertAddressIsPresentOnDisk(self, name: str, address: str, ctx: TemporaryPocketBookRoot):
        address_index_path = os.path.join(ctx.root, AddressBook.INDEX_FILE_NAME)
        self.assertTrue(os.path.isfile(address_index_path))

        with open(address_index_path, 'r') as address_index_file:
            address_index = toml.load(address_index_file)

            self.assertIn(name, address_index)
            self.assertEqual(address_index[name], address)

    def test_no_writing_on_empty_folder(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            self.assertEqual(len(os.listdir(ctx.root)), 0)

    def test_simple_write(self):
        with TemporaryPocketBookRoot() as ctx:
            # add the sample address
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)

            self.assertAddressIsPresentOnDisk('sample', SAMPLE_ADDRESS, ctx)

    def test_check_keys_exists(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)

            self.assertIn('sample', address_book.keys())
            self.assertNotIn('foo', address_book.keys())

    def test_get_items(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)

            address_book_items = list(address_book.items())
            self.assertEqual(len(address_book_items), 1)
            self.assertEqual(address_book_items[0], ('sample', SAMPLE_ADDRESS))

    def test_lookup_of_address(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)

            self.assertEqual(address_book.lookup_address('sample'), SAMPLE_ADDRESS)

    def test_invalid_lookup_of_address(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            with self.assertRaises(RuntimeError):
                address_book.lookup_address('foo-bar')

    def test_throw_on_duplicate_add(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)

            with self.assertRaises(RuntimeError):
                address_book.add('sample', SAMPLE_ADDRESS)

    def test_load_of_addresses(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book1 = AddressBook(root=ctx.root)
            address_book1.add('sample', SAMPLE_ADDRESS)

            address_book2 = AddressBook(root=ctx.root)
            self.assertEqual(set(address_book1.items()), set(address_book2.items()))

    def test_rename(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book1 = AddressBook(root=ctx.root)
            address_book1.add('sample', SAMPLE_ADDRESS)
            self.assertTrue(address_book1.rename('sample', 'sample2'))
            self.assertIn('sample2', address_book1.keys())
            self.assertNotIn('sample', address_book1.keys())

    def test_rename_occurs_on_disk(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book1 = AddressBook(root=ctx.root)
            address_book1.add('sample', SAMPLE_ADDRESS)

            self.assertTrue(address_book1.rename('sample', 'sample2'))
            self.assertAddressIsNotPresentOnDisk('sample', ctx)
            self.assertAddressIsPresentOnDisk('sample2', SAMPLE_ADDRESS, ctx)

    def test_rename_fails(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book1 = AddressBook(root=ctx.root)
            self.assertFalse(address_book1.rename('sample', 'sample2'))

    def test_remove(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)
            self.assertTrue(address_book.remove('sample'))

            self.assertNotIn('sample', address_book.keys())

    def test_remove_is_stored_on_disk(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            address_book.add('sample', SAMPLE_ADDRESS)
            self.assertTrue(address_book.remove('sample'))

            self.assertAddressIsNotPresentOnDisk('sample', ctx)

    def test_removed_failed(self):
        with TemporaryPocketBookRoot() as ctx:
            address_book = AddressBook(root=ctx.root)
            self.assertFalse(address_book.remove('sample'))



import unittest
from unittest.mock import patch, Mock


class AddCommandTests(unittest.TestCase):

    @patch('pocketbook.address_book.AddressBook')
    def test_add(self, address_book):
        from pocketbook.commands.add import run_add

        args = Mock()
        args.name = 'foo'
        args.address = 'foo-address'

        run_add(args)

        instance = address_book.return_value
        instance.add.assert_called_once_with('foo', 'foo-address')



import unittest
from io import StringIO
from unittest.mock import patch, Mock, call

from pocketbook.address_book import AddressBook
from pocketbook.key_store import KeyStore
from pocketbook.table import Table
from pocketbook.utils import create_api, get_balance, get_stake
from tests.utils import SAMPLE_ADDRESS


class ListCommandTests(unittest.TestCase):

    @patch('sys.stdout', new_callable=StringIO)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_empty_list(self, MockKeyStore, MockAddressBook, output, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []

        # run the command
        from pocketbook.commands.list import run_list
        run_list(None)

        self.assertIn('No keys present', output.getvalue())

    @patch('pocketbook.utils.get_stake', spec=get_stake)
    @patch('pocketbook.utils.get_balance', spec=get_balance)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.table.Table', spec=Table)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_normal_display(self, MockKeyStore, MockAddressBook, MockTable, mock_create_api, mock_get_balance,
                            mock_get_stake, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['sample']
        key_store.lookup_address.return_value = SAMPLE_ADDRESS

        address_book = MockAddressBook()
        address_book.items.return_value = [('other', 'another-address')]

        mock_get_balance.side_effect = [10, 5]
        mock_get_stake.side_effect = [5, 10]

        api = mock_create_api()
        mock_create_api.reset_mock()

        args = Mock()
        args.verbose = False
        args.network = 'bar-net'
        args.pattern = ['*']

        table = MockTable()
        MockTable.reset_mock()

        # run the command
        from pocketbook.commands.list import run_list
        run_list(args)

        MockTable.assert_called_once_with(['name', 'type', 'balance', 'stake'])
        mock_create_api.assert_called_once_with('bar-net')
        key_store.lookup_address.assert_called_once_with('sample')

        self.assertEqual(mock_get_balance.call_args_list, [call(api, SAMPLE_ADDRESS), call(api, 'another-address')])
        self.assertEqual(mock_get_stake.call_args_list, [call(api, SAMPLE_ADDRESS), call(api, 'another-address')])

        # check that we call the correct number of calls
        expected_row_calls = [
            call(name='sample', type='key', balance=10, stake=5, address=SAMPLE_ADDRESS),
            call(name='other', type='addr', balance=5, stake=10, address='another-address'),
        ]
        self.assertEqual(table.add_row.call_args_list, expected_row_calls)
        table.display.assert_called_once_with()

    @patch('pocketbook.utils.get_stake', spec=get_stake)
    @patch('pocketbook.utils.get_balance', spec=get_balance)
    @patch('pocketbook.utils.create_api', spec=create_api)
    @patch('pocketbook.table.Table', spec=Table)
    @patch('pocketbook.address_book.AddressBook', spec=AddressBook)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_verbose_display(self, MockKeyStore, MockAddressBook, MockTable, mock_create_api, mock_get_balance,
                             mock_get_stake, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['sample']
        key_store.lookup_address.return_value = SAMPLE_ADDRESS

        address_book = MockAddressBook()
        address_book.items.return_value = [('other', 'another-address')]

        mock_get_balance.side_effect = [10, 5]
        mock_get_stake.side_effect = [5, 10]

        api = mock_create_api()
        mock_create_api.reset_mock()

        args = Mock()
        args.verbose = True
        args.network = 'bar-net'
        args.pattern = ['*']

        table = MockTable()
        MockTable.reset_mock()

        # run the command
        from pocketbook.commands.list import run_list
        run_list(args)

        MockTable.assert_called_once_with(['name', 'type', 'balance', 'stake', 'address'])
        mock_create_api.assert_called_once_with('bar-net')
        key_store.lookup_address.assert_called_once_with('sample')

        self.assertEqual(mock_get_balance.call_args_list, [call(api, SAMPLE_ADDRESS), call(api, 'another-address')])
        self.assertEqual(mock_get_stake.call_args_list, [call(api, SAMPLE_ADDRESS), call(api, 'another-address')])

        # check that we call the correct number of calls
        expected_row_calls = [
            call(name='sample', type='key', balance=10, stake=5, address=SAMPLE_ADDRESS),
            call(name='other', type='addr', balance=5, stake=10, address='another-address'),
        ]
        self.assertEqual(table.add_row.call_args_list, expected_row_calls)
        table.display.assert_called_once_with()



import unittest
from io import StringIO
from unittest.mock import patch

from fetchai.ledger.crypto import Entity

from pocketbook.key_store import KeyStore
from tests.utils import SUPER_SECURE_PASSWORD


class CreateCommandTests(unittest.TestCase):

    @patch('builtins.input', side_effect=['foo-bar'])
    @patch('getpass.getpass', side_effect=[SUPER_SECURE_PASSWORD, SUPER_SECURE_PASSWORD])
    @patch('fetchai.ledger.crypto.Entity', spec=Entity)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_simple_flow(self, MockKeyStore, MockEntity, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []
        entity = MockEntity()

        from pocketbook.commands.create import run_create
        run_create(None)

        key_store.add_key.assert_called_once_with('foo-bar', SUPER_SECURE_PASSWORD, entity)

    @patch('builtins.input', side_effect=['foo-bar'])
    @patch('getpass.getpass', side_effect=[SUPER_SECURE_PASSWORD, '', SUPER_SECURE_PASSWORD, SUPER_SECURE_PASSWORD])
    @patch('sys.stdout', new_callable=StringIO)
    @patch('fetchai.ledger.crypto.Entity', spec=Entity)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_mismatching_passwords(self, MockKeyStore, MockEntity, output, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []
        entity = MockEntity()

        from pocketbook.commands.create import run_create
        run_create(None)

        key_store.add_key.assert_called_once_with('foo-bar', SUPER_SECURE_PASSWORD, entity)

        self.assertIn('Passwords did not match, try again', output.getvalue())

    @patch('builtins.input', side_effect=['foo-bar'])
    @patch('getpass.getpass', side_effect=['weak', SUPER_SECURE_PASSWORD, SUPER_SECURE_PASSWORD])
    @patch('sys.stdout', new_callable=StringIO)
    @patch('fetchai.ledger.crypto.Entity', spec=Entity)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_weak_password(self, MockKeyStore, MockEntity, output, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = []
        entity = MockEntity()

        from pocketbook.commands.create import run_create
        run_create(None)

        key_store.add_key.assert_called_once_with('foo-bar', SUPER_SECURE_PASSWORD, entity)

        self.assertIn('Password too simple, try again', output.getvalue())

    @patch('builtins.input', side_effect=['foo-bar', 'foo-baz'])
    @patch('getpass.getpass', side_effect=[SUPER_SECURE_PASSWORD, SUPER_SECURE_PASSWORD])
    @patch('sys.stdout', new_callable=StringIO)
    @patch('fetchai.ledger.crypto.Entity', spec=Entity)
    @patch('pocketbook.key_store.KeyStore', spec=KeyStore)
    def test_re_prompt_for_duplicate_key(self, MockKeyStore, MockEntity, output, *args):
        key_store = MockKeyStore()
        key_store.list_keys.return_value = ['foo-bar']
        entity = MockEntity()

        from pocketbook.commands.create import run_create
        run_create(None)

        key_store.add_key.assert_called_once_with('foo-baz', SUPER_SECURE_PASSWORD, entity)

        self.assertIn('Key name already exists', output.getvalue())



import os

from .constants import DEFAULT_KEY_STORE_ROOT

DISCLAIMER = """
                                == Warning ==

You use this application at your own risk. Whilst Fetch.ai have made every
effort to ensure its reliability and security, it comes with no warranty. It is
intended for the creation and management of Fetch.ai mainnet wallets and
transactions between them. You are responsible for the security of your own
private keys (see ~/.pocketbook folder). Do not use this application for
high-value operations: it is intended for utility operations on the main network.
"""


def display_disclaimer(root=None):
    key_store_root = root or DEFAULT_KEY_STORE_ROOT
    if not os.path.isdir(key_store_root):
        print(DISCLAIMER)
        input('Press enter to accept')



def run_rename(args):
    from pocketbook.address_book import AddressBook
    from pocketbook.key_store import KeyStore

    address_book = AddressBook()
    key_store = KeyStore()

    # make sure that the new name is not present either as a key, or as an address
    new_present = args.new in address_book.keys() or args.new in key_store.list_keys()
    if new_present:
        print('{} is already present, please choose a different destination name'.format(args.new))
        return 1

    # check the old address or key name
    old_is_address = args.old in address_book.keys()
    old_is_key = args.old in key_store.list_keys()

    success = False
    if old_is_address and old_is_key:
        raise RuntimeError('Data store corrupting, key looks like an address + key')
    elif old_is_address:
        success = address_book.rename(args.old, args.new)
    elif old_is_key:
        success = key_store.rename_key(args.old, args.new)
    else:
        print('{} doesn\'t appear to be a valid key or address name, please check and try again'.format(args.old))
        return 1

    if not success:
        print('Failed to rename {} to {}'.format(args.old, args.new))
        return 1

    return 0



def run_transfer(args):
    from getpass import getpass

    from fetchai.ledger.crypto import Address
    from fetchai.ledger.serialisation.transaction import encode_transaction

    from pocketbook.address_book import AddressBook
    from pocketbook.key_store import KeyStore
    from pocketbook.utils import create_api, from_canonical, token_amount

    address_book = AddressBook()
    key_store = KeyStore()

    # choose the destination
    destination_name = '{}:'.format(args.destination)
    if args.destination in address_book.keys():
        destination = address_book.lookup_address(args.destination)
    else:
        destination = key_store.lookup_address(args.destination)
        if destination is None:
            destination = Address(args.destination)
            destination_name = ''

    # convert the amount
    amount = args.amount
    charge_rate = args.charge_rate
    computed_amount = from_canonical(amount)

    # check all the signers make sense
    for signer in args.signers:
        if signer not in key_store.list_keys():
            raise RuntimeError('Unknown key: {}'.format(signer))

    # determine the from account
    from_address_name = None
    if len(args.signers) == 1 and args.from_address is None:
        from_address_name = args.signers[0]
    elif len(args.signers) >= 1 and args.from_address is not None:
        present = args.from_address in key_store.list_keys() or args.from_address in address_book.keys()
        from_address_name = args.from_address
        if not present:
            raise RuntimeError('Unknown from address: {}'.format(args.from_address))
    else:
        raise RuntimeError('Unable to determine from account')

    required_ops = len(args.signers)
    fee = required_ops * charge_rate
    computed_fee = from_canonical(fee)
    computed_total = computed_amount + computed_fee
    computed_charge_rate = from_canonical(charge_rate)

    print('Network....:', args.network)
    print('From.......:', str(from_address_name))
    print('Signer(s)..:', ','.join(args.signers))
    print('Destination:', destination_name, str(destination))
    print('Amount.....:', token_amount(computed_amount))
    print('Fee........:', token_amount(computed_fee))

    # only display extended fee information if something other than the default it selected
    if charge_rate != 1:
        print('           : {} ops @ {}'.format(required_ops, token_amount(computed_charge_rate)))

    print('Total......:', token_amount(computed_total), '(Amount + Fee)')
    print()
    input('Press enter to continue')

    api = create_api(args.network)

    # start unsealing the private keys
    entities = {}
    for signer in args.signers:
        entity = key_store.load_key(signer, getpass('Enter password for key {}: '.format(signer)))
        entities[signer] = entity

    from_address = None
    if from_address_name in entities:
        from_address = Address(entities[from_address_name])
    elif from_address_name in address_book.keys():
        from_address = Address(address_book.lookup_address(from_address_name))

    # build up the basic transaction information
    tx = api.tokens._create_skeleton_tx(required_ops)
    tx.charge_rate = charge_rate
    tx.from_address = Address(from_address)
    tx.add_transfer(destination, amount)
    for entity in entities.values():
        tx.add_signer(entity)

    # encode and sign the transaction
    encoded_tx = encode_transaction(tx, list(entities.values()))

    # # submit the transaction
    print('Submitting TX...')
    api.sync(api.tokens._post_tx_json(encoded_tx, 'transfer'))
    print('Submitting TX...complete')



from io import open
from os import path

from setuptools import setup, find_packages

from pocketbook import __version__

project_root = path.dirname(__file__)

with open(path.join(project_root, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pocketbook',
    version=__version__,
    description='Command line wallet application for the Fetch.ai network',
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',
    url='https://github.com/fetchai/tools-pocketbook',
    author='Edward FitzGerald',
    author_email='edward.fitzgerald@fetch.ai',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    python_requires='~=3.5',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'fetchai-ledger-api==1.0.0rc1',
        'toml',
        'colored',
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage', 'pytest', 'flake8'],
    },
    entry_points={
        'console_scripts': [
            'pocketbook=pocketbook.cli:main'
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/fetchai/tools-pocketbook/issues',
        'Source': 'https://github.com/fetchai/tools-pocketbook',
    },
)
